import os
import glob
import time
import logging
from rich.logging import RichHandler
import getpass
from typing import Union
from .local import submit_local, submit_venv, submit_conda
from .ssh import submit_ssh, monitor_ssh
from .cluster import submit_sge, monitor_sge, submit_slurm, monitor_slurm
from .cloud import submit_gcp, monitor_gcp, clean_up_gcp


# Overview of implemented remote resources in addition to local processes
cluster_resources = ["sge-cluster", "slurm-cluster"]
cloud_resources = ["gcp-cloud"]
other_resources = ["ssh-node", "local"]


class MLEJob(object):
    """
    Basic Job Class - Everything builds on top of this!

    This class defines, executes & monitors an individual job that can
    either be run locally, on a sungrid-engine (SGE) or SLURM cluster.

    Args:
        resource_to_run (str): Compute resource to execute job on.

        job_filename (str): filepath to .py script to be executed for job.

        config_filename (str): filepath to .json file specifying configuration
            of experiment to be executed in this job.

        job_arguments (dict): ressources required for this specific job.

        experiment_dir (str): set path for logging & unique storage of results.

        cmd_line_input (dict): provides standardized cmd input to .py file in
            job submission. Includes -config, -exp_dir, -seed.

        extra_cmd_line_input (dict): provides additional cmd input .py file
            in job submission. Dictionary should be structured so that input
            are passed as -<key> <value>.

        logger_level (str): control logger verbosity of individual experiment.

    Methods:
        run: Executes job, logs it & returns status if job done
        schedule: Schedules job locally or remotely
        schedule_local: Schedules job locally on your machine
        schedule_cluster: Schedules job remotely on SGE/Slurm clusters
        monitor: Monitors job locally or remotely
        monitor_local: Monitors job locally on your machine
        monitor_cluster: Monitors job remotely on SGE/Slurm clusters
    """

    def __init__(
        self,
        resource_to_run: str,
        job_filename: str,
        job_arguments: dict = {},
        config_filename: Union[None, str] = None,
        experiment_dir: Union[None, str] = None,
        seed_id: Union[None, int] = None,
        extra_cmd_line_input: Union[None, dict] = None,
        delete_config: bool = False,
        debug_mode: bool = False,
        cloud_settings: Union[dict, None] = None,
        ssh_settings: Union[dict, None] = None,
        logger_level: int = logging.WARNING,
    ):
        # Init job class with relevant info
        self.resource_to_run = resource_to_run  # compute resource for job
        self.job_filename = job_filename  # path to train script
        self.config_filename = config_filename  # path to config json
        self.job_arguments = job_arguments.copy()  # Job resource configuration
        self.experiment_dir = experiment_dir  # main results dir (create)
        self.seed_id = seed_id  # random seed to be passed as cmd-line arg
        self.delete_config = delete_config  # Option to delete config file after run
        self.debug_mode = debug_mode  # Pipe stdout and stderr to files
        self.user_name = getpass.getuser()

        # Create command line arguments for job to schedule (passed to .py)
        self.cmd_line_args = self.generate_cmd_line_args()
        # Add additional cmd line args if extra ones are specified
        if extra_cmd_line_input is not None:
            self.cmd_line_args = self.generate_extra_cmd_line_args(
                self.cmd_line_args, extra_cmd_line_input.copy()
            )

        # GCP/SSH configurations
        if self.resource_to_run in cloud_resources:
            assert cloud_settings is not None
        self.cloud_settings = cloud_settings

        if self.resource_to_run == "ssh-node":
            assert ssh_settings is not None
        self.ssh_settings = ssh_settings

        # Instantiate/connect a logger
        FORMAT = "%(message)s"
        logging.basicConfig(
            level=logger_level,
            format=FORMAT,
            datefmt="[%X]",
            handlers=[RichHandler()],
        )

        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logger_level)

    def run(self) -> bool:
        """Schedule experiment, monitor and clean up afterwards."""
        # Schedule job, create success boolean and job identifier
        job_id = self.schedule()
        # Monitor status of job based on identifier
        status_out = self.monitor(job_id, continuous=True)
        # If they exist - remove log & error file
        if not self.debug_mode:
            self.clean_up(job_id)
        return status_out

    def schedule(self) -> str:
        """Schedule job on cluster (sge/slurm), cloud (gcp) or locally."""
        if self.resource_to_run in cluster_resources:
            job_id = self.schedule_cluster()
            if self.job_status == 1:
                self.logger.info(
                    f"Job ID: {job_id} - Cluster job scheduled"
                    f" - {self.config_filename}"
                )
            else:
                self.logger.info(
                    f"Job ID: {job_id} - Error when scheduling "
                    f"cluster job - {self.config_filename}"
                )
        elif self.resource_to_run in cloud_resources:
            job_id = self.schedule_cloud()
            if self.job_status == 1:
                self.logger.info(
                    f"VM Name: {job_id} - Cloud job scheduled"
                    f" - {self.config_filename}"
                )
            else:
                self.logger.info(
                    f"VM Name: {job_id} - Error when scheduling "
                    f"cloud job - {self.config_filename}"
                )
        elif self.resource_to_run == "local":
            job_id = self.schedule_local()
            if self.job_status == 1:
                self.logger.info(
                    f"PID: {job_id.pid} - Local job scheduled "
                    f"- {self.config_filename}"
                )
            else:
                self.logger.info(
                    f"PID: {job_id.pid} - Error when scheduling "
                    f"local job - {self.config_filename}"
                )
        elif self.resource_to_run == "ssh-node":
            job_id = self.schedule_ssh()
            if self.job_status == 1:
                self.logger.info(
                    f"PID: {job_id} - SSH job scheduled " f"- {self.config_filename}"
                )
            else:
                self.logger.info(
                    f"PID: {job_id.pid} - Error when scheduling "
                    f"SSH job - {self.config_filename}"
                )
        else:
            raise ValueError(f"{self.resource_to_run} is not implemented.")
        # Return sge/slurm - job_id (qstat/squeue), gcp - vm_name, ssh/local - proc
        return job_id

    def monitor(self, job_id: str, continuous: bool = True) -> bool:
        """
        Monitor on cluster (sge/slurm), cloud (gcp) or locally via id.
        If 'continuous' bool is true, then monitor until experiment status 0.
        """
        if self.resource_to_run in cluster_resources:
            if continuous:
                self.logger.info(
                    f"Job ID: {job_id} - Started monitoring -"
                    f" {self.config_filename}"
                )
            status_out = self.monitor_cluster(job_id, continuous)
            if status_out == 0:
                self.logger.info(
                    f"Job ID: {job_id} - Cluster job completed -"
                    f" {self.config_filename}"
                )
        elif self.resource_to_run in cloud_resources:
            if continuous:
                self.logger.info(
                    f"VM Name: {job_id} - Started monitoring -"
                    f" {self.config_filename}"
                )
            status_out = self.monitor_cloud(job_id, continuous)
            if status_out == 0:
                self.logger.info(
                    f"VM Name: {job_id} - Cloud job completed -"
                    f" {self.config_filename}"
                )
        elif self.resource_to_run == "ssh-node":
            if continuous:
                self.logger.info(
                    f"SSH PID: {job_id} - Started monitoring -"
                    f" {self.config_filename}"
                )
            status_out = self.monitor_ssh(job_id, continuous)
            if status_out == 0:
                self.logger.info(
                    f"SSH PID: {job_id} - SSH job completed -"
                    f" {self.config_filename}"
                )
        elif self.resource_to_run == "local":
            if continuous:
                self.logger.info(
                    f"PID: {job_id.pid} - Started monitoring -"
                    f" {self.config_filename}"
                )
            status_out = self.monitor_local(job_id, continuous)
            if status_out == 0:
                self.logger.info(
                    f"PID: {job_id.pid} - Local job completed -"
                    f" { self.config_filename}"
                )
        return status_out

    def schedule_local(self):
        """Schedules job locally on your machine."""
        if "use_conda_venv" in self.job_arguments.keys():
            if self.job_arguments["use_conda_venv"]:
                proc = submit_conda(
                    self.job_filename,
                    self.cmd_line_args,
                    self.job_arguments,
                    self.debug_mode,
                )
            else:
                proc = submit_local(
                    self.job_filename, self.cmd_line_args, self.debug_mode
                )
        elif "use_venv_venv" in self.job_arguments.keys():
            if self.job_arguments["use_venv_venv"]:
                proc = submit_venv(
                    self.job_filename,
                    self.cmd_line_args,
                    self.job_arguments,
                    self.debug_mode,
                )
            else:
                proc = submit_local(
                    self.job_filename, self.cmd_line_args, self.debug_mode
                )
        else:
            proc = submit_local(self.job_filename, self.cmd_line_args, self.debug_mode)
        self.job_status = 1
        return proc

    def schedule_ssh(self):
        """Schedules job on SSH servers."""
        proc = submit_ssh(
            self.job_filename,
            self.cmd_line_args,
            self.job_arguments,
            self.ssh_settings,
            self.debug_mode,
        )
        self.job_status = 1
        return proc

    def schedule_cluster(self) -> int:
        """Schedules job to run remotely on SGE or Slurm clusters."""
        if self.resource_to_run == "sge-cluster":
            job_id = submit_sge(
                self.job_filename,
                self.cmd_line_args,
                self.job_arguments,
                self.user_name,
                self.debug_mode,
                clean_up=True,
            )
        elif self.resource_to_run == "slurm-cluster":
            job_id = submit_slurm(
                self.job_filename,
                self.cmd_line_args,
                self.job_arguments,
                self.user_name,
                self.debug_mode,
                clean_up=True,
            )
        if job_id == -1:
            self.job_status = 0
        else:
            self.job_status = 1
        return job_id

    def schedule_cloud(self) -> int:
        """Schedules job to run remotely on GCP cloud."""
        if self.resource_to_run == "gcp-cloud":
            # Submit VM Creation + Startup exec
            job_id = submit_gcp(
                self.job_filename,
                self.cmd_line_args,
                self.experiment_dir,
                self.job_arguments,
                self.debug_mode,
                self.cloud_settings,
            )
        if job_id == -1:
            self.job_status = 0
        else:
            self.job_status = 1
        return job_id

    def monitor_local(self, proc, continuous: bool = True) -> int:
        """Monitors job locally on your machine."""
        # Poll status of local process & change status when done
        if continuous:
            while self.job_status:
                poll = proc.poll()
                if poll is None:
                    continue
                else:
                    self.job_status = 0
                # Sleep until next status check
                time.sleep(1)

            # Get output & error messages (if there is an error)
            out, err = proc.communicate()
            # Return -1 if job failed & 0 otherwise
            if proc.returncode != 0:
                print(out, err)
                return -1
            else:
                return 0
        else:
            poll = proc.poll()

            # Get output & error messages (if there is an error)
            out, err = proc.communicate()
            if poll is None:
                return 1
            else:
                return 0

    def monitor_ssh(self, proc, continuous: bool = True) -> int:
        """Monitors job remotely on SSH server."""
        # Poll status of local process & change status when done
        if continuous:
            while self.job_status:
                self.job_status = monitor_ssh(proc, self.ssh_settings)
                # Sleep until next status check
                time.sleep(1)
            return 0
        else:
            return monitor_ssh(proc, self.ssh_settings)

    def monitor_cluster(self, job_id: str, continuous: bool = True) -> int:
        """Monitors job remotely on SGE or Slurm clusters."""
        if continuous:
            while self.job_status:
                if self.resource_to_run == "sge-cluster":
                    self.job_status = monitor_sge(job_id, self.user_name)
                elif self.resource_to_run == "slurm-cluster":
                    self.job_status = monitor_slurm(job_id, self.user_name)
                time.sleep(1)
            return 0
        else:
            if self.resource_to_run == "sge-cluster":
                return monitor_sge(job_id, self.user_name)
            elif self.resource_to_run == "slurm-cluster":
                return monitor_slurm(job_id, self.user_name)

    def monitor_cloud(self, job_id: str, continuous: bool = True) -> int:
        """Monitors job remotely on GCP cloud."""
        if continuous:
            while self.job_status:
                if self.resource_to_run == "gcp-cloud":
                    self.job_status = monitor_gcp(job_id, self.job_arguments)
                time.sleep(10)
            return 0
        else:
            if self.resource_to_run == "gcp-cloud":
                return monitor_gcp(job_id, self.job_arguments)

    def clean_up(self, job_id: str) -> None:
        """Remove error and log files at end of training."""
        if self.resource_to_run in cluster_resources:
            for filename in glob.glob("err.err"):
                try:
                    os.remove(filename)
                except Exception:
                    pass
            for filename in glob.glob("log.txt"):
                try:
                    os.remove(filename)
                except Exception:
                    pass
            self.logger.info("Cleaned up log, error, results files")

        # Delete VM instance and code directory stored in data bucket
        if self.resource_to_run == "gcp-cloud":
            clean_up_gcp(
                job_id,
                self.job_arguments,
                self.experiment_dir,
                self.cloud_settings,
            )
            self.logger.info(f"VM Name: {job_id} - Delete VM - {self.config_filename}")

        # If desired also delete configuration files
        if self.delete_config:
            try:
                os.remove(self.config_filename)
            except Exception:
                pass

    def generate_cmd_line_args(self) -> str:
        """Generate cmd line args for .py -> get_train_configs_ready"""
        cmd_line_args = ""
        if self.experiment_dir is not None:
            cmd_line_args += " -exp_dir " + self.experiment_dir

        if self.config_filename is not None:
            cmd_line_args += " -config " + self.config_filename

        if self.seed_id is not None:
            cmd_line_args += " -seed " + str(self.seed_id)
            # Update the job argument details with the seed-job-id
            if "job_name" in self.job_arguments.keys():
                self.job_arguments["job_name"] += "-" + str(self.seed_id)
            else:
                self.job_arguments["job_name"] = str(self.seed_id)
        return cmd_line_args

    def generate_extra_cmd_line_args(
        self, cmd_line_args: str, extra_cmd_line_input: Union[None, dict] = None
    ) -> str:
        """Generate extra cmd line args for .py -> e.g. for postproc"""
        full_cmd_line_args = (cmd_line_args + ".")[:-1]
        if extra_cmd_line_input is not None:
            for k, v in extra_cmd_line_input.items():
                full_cmd_line_args += " -" + k + " " + str(v)
        return full_cmd_line_args



import logging
from typing import List


class SSH_Manager(object):
    """SSH client for file transfer & local 2 remote experiment exec."""

    def __init__(
        self,
        user_name: str,
        pkey_path: str,
        main_server: str,
        jump_server: str = "",
        ssh_port: int = 22,
    ):
        """Set the credentials & resource details."""
        self.pkey_path = pkey_path
        self.main_server = main_server
        self.jump_server = jump_server
        self.port = ssh_port
        self.user = user_name

        # We are always using tunnel even if not necessary!
        if self.jump_server == "":
            self.jump_server = self.main_server

    def generate_tunnel(self):
        """Generate a tunnel through the jump host."""
        try:
            from sshtunnel import SSHTunnelForwarder

            logger = logging.getLogger("sshtunnel")
            logger.propagate = False
            logger.setLevel(logging.CRITICAL + 1)
        except ImportError:
            raise ImportError(
                "You need to install `sshtunnel` to use SSH forwarding."
            )
        return SSHTunnelForwarder(
            (self.jump_server, self.port),
            ssh_username=self.user,
            ssh_password="",
            ssh_pkey=self.pkey_path,
            remote_bind_address=(self.main_server, self.port),
        )

    def connect(self, tunnel):
        """Connect to the ssh client."""
        try:
            import paramiko

            logger = logging.getLogger("paramiko")
            logger.setLevel(logging.ERROR)
        except ImportError:
            raise ImportError(
                "You need to install `paramiko` to use the SSH client."
            )

        while True:
            try:
                client = paramiko.SSHClient()
                client.load_system_host_keys()
                client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
                client.connect(
                    hostname=tunnel.local_bind_host,
                    port=tunnel.local_bind_port,
                    username=self.user,
                    password="",
                    timeout=100,
                )
                break
            except Exception:
                continue
        return client

    def sync_dir(self, local_dir_name, remote_dir_name):
        """Clone/sync over a local directory to remote server."""
        try:
            from scp import SCPClient

        except ImportError:
            raise ImportError(
                "You need to install `scp` to use the SCP client."
            )

        with self.generate_tunnel() as tunnel:
            client = self.connect(tunnel)
            scp = SCPClient(client.get_transport())
            scp.put(local_dir_name, remote_dir_name, recursive=True)
            client.close()
        return

    def execute_command(self, cmds_to_exec: List[str]):
        """Execute a shell command on the remote server."""
        for cmd in cmds_to_exec:
            with self.generate_tunnel() as tunnel:
                client = self.connect(tunnel)
                stdin, stdout, stderr = client.exec_command(cmd, get_pty=True)
                for line in stderr:
                    print(line)
                client.close()
        return stdin, stdout, stderr

    def read_file(self, file_name: str):
        """Read a file from remote server and return it."""
        with self.generate_tunnel() as tunnel:
            client = self.connect(tunnel)
            ftp = client.open_sftp()
            # Something doesnt work here
            remote_file = ftp.open(file_name)
            all_lines = []
            try:
                for line in list(remote_file):
                    all_lines.append(line)
            except Exception as e:
                print(e)
            ftp.close()
            client.close()
        return all_lines

    def write_to_file(self, str_to_write: str, file_name: str):
        """Write a string to a text file on remote server."""
        with self.generate_tunnel() as tunnel:
            client = self.connect(tunnel)
            ftp = client.open_sftp()
            file = ftp.file(file_name, "w", -1)
            file.write(str_to_write)
            file.flush()
            ftp.close()
            client.close()
        return

    def get_file(self, remote_dir_name: str, local_dir_name: str):
        """scp clone a remote directory to the local machine."""
        try:
            from scp import SCPClient

        except ImportError:
            raise ImportError(
                "You need to install `scp` to use the SCP client."
            )

        with self.generate_tunnel() as tunnel:
            client = self.connect(tunnel)
            scp = SCPClient(client.get_transport())
            scp.get(remote_dir_name, local_path=local_dir_name, recursive=True)
            client.close()
        return

    def delete_file(self, file_name: str):
        """Delete a file on the remote server."""
        with self.generate_tunnel() as tunnel:
            client = self.connect(tunnel)
            ftp = client.open_sftp()
            ftp.remove(file_name)
            ftp.close()
            client.close()
        return

    def delete_dir(self, file_name: str):
        """Delete a directory on the remote server."""
        cmd = "rm -rf " + file_name
        self.execute_command([cmd])



from .manage_local import (
    submit_venv,
    submit_conda,
    submit_local,
    submit_subprocess,
    random_id,
)


__all__ = [
    "submit_venv",
    "submit_conda",
    "submit_local",
    "submit_subprocess",
    "random_id",
]



import os
import glob
import logging
from typing import Union


def send_dir_gcp(
    cloud_settings: dict,
    local_dir: Union[str, None] = None,
    number_of_connect_tries: int = 5,
):
    """Send entire dir (recursively) to Google Cloud Storage Bucket."""
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)
    try:
        from google.cloud import storage

    except ImportError:
        raise ImportError(
            "You need to install `google-cloud-storage` to use GCP buckets."
        )

    for i in range(number_of_connect_tries):
        try:
            client = storage.Client(cloud_settings["project_name"])
            bucket = client.get_bucket(
                cloud_settings["bucket_name"], timeout=20
            )
        except Exception:
            logger.info(
                f"Attempt {i+1}/{number_of_connect_tries}"
                " - Failed sending to GCloud Storage"
            )

    def upload_single_file(local_path, gcs_path, bucket):
        # Recursively upload the folder structure
        if os.path.isdir(local_path):
            for local_file in glob.glob(os.path.join(local_path, "**")):
                if not os.path.isfile(local_file):
                    upload_single_file(
                        local_file,
                        os.path.join(gcs_path, os.path.basename(local_file)),
                        bucket,
                    )
                else:
                    remote_path = os.path.join(
                        gcs_path, local_file[1 + len(local_path) :]
                    )
                    blob = bucket.blob(remote_path)
                    blob.upload_from_filename(local_file)
        # Only upload single file - e.g. zip compressed experiment
        else:
            remote_path = gcs_path
            blob = bucket.blob(remote_path)
            blob.upload_from_filename(local_path)

    if local_dir is None:
        local_dir = os.getcwd()
    upload_single_file(local_dir, cloud_settings["remote_dir"], bucket)


def copy_dir_gcp(
    cloud_settings: dict,
    remote_dir: str,
    local_dir: Union[str, None] = None,
    number_of_connect_tries: int = 5,
):
    """Download entire dir (recursively) from Google Cloud Storage Bucket."""
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)
    try:
        from google.cloud import storage

    except ImportError:
        raise ImportError(
            "You need to install `google-cloud-storage` to use GCP buckets."
        )

    for i in range(number_of_connect_tries):
        try:
            client = storage.Client(cloud_settings["project_name"])
            bucket = client.get_bucket(
                cloud_settings["bucket_name"], timeout=20
            )
        except Exception:
            logger.info(
                f"Attempt {i+1}/{number_of_connect_tries}"
                " - Failed sending to GCloud Storage"
            )

    blobs = bucket.list_blobs(prefix=remote_dir)  # Get list of files
    blobs = list(blobs)

    if local_dir is None:
        path = os.path.normpath(remote_dir)
        local_dir = path.split(os.sep)[-1]

    # Recursively download the folder structure
    if len(blobs) > 1:
        local_path = os.path.expanduser(local_dir)
        for blob in blobs:
            filename = blob.name[len(remote_dir) :]
            local_f_path = os.path.join(local_path, filename)
            dir_path = os.path.dirname(local_f_path)

            if not os.path.exists(dir_path):
                try:
                    os.makedirs(dir_path)
                except Exception:
                    pass
            temp_file = local_path + filename
            temp_dir = os.path.split(temp_file)[0]
            if not os.path.exists(temp_dir):
                os.makedirs(temp_dir)
            blob.download_to_filename(temp_file)
    # Only download single file - e.g. zip compressed experiment
    else:
        for blob in blobs:
            filename = blob.name.split("/")[1]
            blob.download_to_filename(filename)


def delete_dir_gcp(
    cloud_settings: dict,
    number_of_connect_tries: int = 5,
):
    """Delete a directory in a GCS bucket."""
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)
    try:
        from google.cloud import storage

    except ImportError:
        raise ImportError(
            "You need to install `google-cloud-storage` to use GCP buckets."
        )

    for i in range(number_of_connect_tries):
        try:
            client = storage.Client(cloud_settings["project_name"])
            bucket = client.get_bucket(
                cloud_settings["bucket_name"], timeout=20
            )
        except Exception:
            logger.info(
                f"Attempt {i+1}/{number_of_connect_tries}"
                " - Failed sending to GCloud Storage"
            )

    # Delete all files in directory
    blobs = bucket.list_blobs(prefix=cloud_settings["remote_dir"])
    for blob in blobs:
        try:
            blob.delete()
        except Exception:
            pass



import os
import shutil
from mle_scheduler import MLEJob, MLEQueue

# Only test this locally! For remote - have to run example `run_XXX.py` manually


def test_job():
    # Launch a python train_mnist.py -config base_config.json job
    job = MLEJob(
        resource_to_run="local",
        job_filename="examples/train.py",
        config_filename="examples/base_config_1.yaml",
        experiment_dir="logs_single",
        job_arguments={},
    )

    job.run()

    # Check existence of expected files
    assert os.path.exists(
        os.path.join("logs_single", "base_config_1", "logs/log_seed_1.hdf5")
    )
    shutil.rmtree("logs_single")
    return


def test_queue():
    # Launch a queue of 4 jobs (2 configs x 2 seeds)
    queue = MLEQueue(
        resource_to_run="local",
        job_filename="examples/train.py",
        config_filenames=[
            "examples/base_config_1.yaml",
            "examples/base_config_2.yaml",
        ],
        num_seeds=2,
        random_seeds=[0, 1],
        experiment_dir="logs_queue",
        job_arguments={},
    )
    queue.run()

    # Check existence of expected files
    assert os.path.exists(
        os.path.join("logs_queue", "base_config_1/logs/log_seed_0.hdf5")
    )
    assert os.path.exists(
        os.path.join("logs_queue", "base_config_1/logs/log_seed_1.hdf5")
    )

    assert os.path.exists(
        os.path.join("logs_queue", "base_config_2/logs/log_seed_0.hdf5")
    )
    assert os.path.exists(
        os.path.join("logs_queue", "base_config_2/logs/log_seed_1.hdf5")
    )

    shutil.rmtree("logs_queue")
    return


def test_automerge_seeds():
    # Launch a queue of 4 jobs (2 configs x 2 seeds)
    queue = MLEQueue(
        resource_to_run="local",
        job_filename="examples/train.py",
        config_filenames=[
            "examples/base_config_1.yaml",
            "examples/base_config_2.yaml",
        ],
        num_seeds=2,
        random_seeds=[0, 1],
        experiment_dir="logs_merge",
        job_arguments={},
        automerge_seeds=True,
    )
    queue.run()

    # Check existence of expected files
    assert os.path.exists(
        os.path.join("logs_merge", "base_config_1/logs/log.hdf5")
    )
    assert os.path.exists(
        os.path.join("logs_merge", "base_config_2/logs/log.hdf5")
    )
    shutil.rmtree("logs_merge")
    return


def test_automerge_configs():
    # Launch a queue of 4 jobs (2 configs x 2 seeds)
    queue = MLEQueue(
        resource_to_run="local",
        job_filename="examples/train.py",
        config_filenames=[
            "examples/base_config_1.yaml",
            "examples/base_config_2.yaml",
        ],
        num_seeds=2,
        random_seeds=[0, 1],
        experiment_dir="logs_merge",
        job_arguments={},
        automerge_configs=True,
    )
    queue.run()

    # Check existence of expected files
    assert os.path.exists(os.path.join("logs_merge", "meta_log.hdf5"))
    assert os.path.exists(
        os.path.join("logs_merge", "base_config_1/logs/log.hdf5")
    )
    assert os.path.exists(
        os.path.join("logs_merge", "base_config_2/logs/log.hdf5")
    )
    shutil.rmtree("logs_merge")
    return



from .startup_script_sge import (
    sge_base_job_config,
    sge_job_exec,
    sge_conda_activate,
    sge_venv_activate,
)


def sge_generate_startup_file(job_arguments: dict) -> str:
    """Generate the bash script template to submit with qsub."""
    # Set the job template depending on the desired number of GPUs
    base_template = (sge_base_job_config + ".")[:-1]

    # Add desired number of requested gpus
    if "num_gpus" in job_arguments:
        if job_arguments["num_gpus"] > 0:
            base_template += '#$ -l {gpu_prefix}="{num_gpus}"'
        if "gpu_type" in job_arguments.keys():
            base_template += ',gputype="{gpu_type}"'
        base_template += "\n"

    # Exclude specific nodes from the queue
    if "exclude_nodes" in job_arguments:
        base_template += (
            "#$ -l hostname="
            + "&".join(("!" + f"{x}" for x in job_arguments["exclude_nodes"]))
            + "\n"
        )

    # Only run on specific nodes from the queue
    if "include_nodes" in job_arguments:
        base_template += (
            "#$ -l hostname="
            + "&".join((f"{x}" for x in job_arguments["include_nodes"]))
            + "\n"
        )

    # Set the max required memory per job in MB - standardization Slurm
    if "memory_per_job" in job_arguments:
        base_template += "#$ -l h_vmem={memory_per_job}M\n"
        base_template += "#$ -l mem_free={memory_per_job}M\n"

    # Set the max required time per job (afterwards terminate)
    if "time_per_job" in job_arguments:
        base_template += "#$ -l h_rt={time_per_job}\n"

    # Add the return of the job id
    base_template += "#$ -terse\n"

    # Add the 'tail' - script execution to the string
    template_out = base_template
    if "use_conda_venv" in job_arguments.keys():
        if job_arguments["use_conda_venv"]:
            template_out += sge_conda_activate
    elif "use_venv_venv" in job_arguments.keys():
        if job_arguments["use_venv_venv"]:
            template_out += sge_venv_activate
    template_out += sge_job_exec
    return template_out



import os
import random
import string
import shlex
import subprocess as sp
import sys

PYTHON_EXECUTABLE = (
    sys.executable
)  # local executable Python file (more reliable than `python`)


def submit_conda(
    filename: str,
    cmd_line_arguments: str,
    job_arguments: dict,
    debug: bool = True,
):
    """Create a local job & submit it based on provided file to execute."""
    f_name, f_extension = os.path.splitext(filename)
    if f_extension == ".py":
        cmd = f"{PYTHON_EXECUTABLE} {filename} {cmd_line_arguments}"
    elif f_extension == ".sh":
        cmd = f"bash {filename} {cmd_line_arguments}"
    else:
        raise ValueError(
            f"Script with {f_extension} cannot be handled"
            " by mle-launcher. Only base .py, .sh experiments"
            " are so far implemented. Please open an issue."
        )

    try:
        env_name = job_arguments["env_name"]
        current_env = os.environ["CONDA_PREFIX"]
        if current_env != env_name:
            # activate the correct conda environment
            cmd = (
                "source $(conda info --base)/etc/profile.d/conda.sh           "
                f"          && conda activate {env_name}                     &&"
                f" {cmd}"
            )
    except Exception:
        pass
    proc = submit_subprocess(cmd, debug)
    return proc


def submit_venv(
    filename: str,
    cmd_line_arguments: str,
    job_arguments: dict,
    debug: bool = True,
):
    """Create a local job & submit it based on provided file to execute."""
    cmd = f"{PYTHON_EXECUTABLE} {filename} {cmd_line_arguments}"
    env_name = job_arguments["env_name"]
    command_template = '/bin/bash -c "source {}/{}/bin/activate && {}"'
    cmd = shlex.split(
        command_template.format(os.environ["WORKON_HOME"], env_name, cmd)
    )
    proc = submit_subprocess(cmd, debug)
    return proc


def submit_local(filename: str, cmd_line_arguments: str, debug: bool = True):
    """Create a local job & submit it based on provided file to execute."""
    cmd = f"{PYTHON_EXECUTABLE} {filename} {cmd_line_arguments}"
    proc = submit_subprocess(cmd, debug)
    return proc


def submit_subprocess(cmd: str, debug: bool = True) -> sp.Popen:
    """Submit a subprocess & return the process."""
    # Pipe stdout & stderr to separate files.
    # if debug:
    #     cmd += " 2>>err.err 1>>log.txt"
    while True:
        try:
            proc = sp.Popen(cmd, shell=True, stdout=sp.PIPE, stderr=sp.PIPE)
            break
        except Exception:
            continue
    return proc


def random_id(length: int = 8) -> str:
    """Sample a random string to use for job id."""
    return "".join(random.sample(string.ascii_letters + string.digits, length))



import os
import time
from typing import Union
from .ssh_manager import SSH_Manager
from .helpers_launch_ssh import ssh_get_submission_cmd


def submit_ssh(
    filename: str,
    cmd_line_arguments: Union[str, None],
    job_arguments: dict,
    ssh_settings: dict,
):
    """Launch a job on an SSH server."""
    # Create bash script string
    script_cmd = ssh_get_submission_cmd(
        filename, cmd_line_arguments, job_arguments, ssh_settings
    )

    # Create manager for file sync and subprocess exec
    ssh_manager = SSH_Manager(
        user_name=ssh_settings["user_name"],
        pkey_path=ssh_settings["pkey_path"],
        main_server=ssh_settings["main_server"],
        jump_server=ssh_settings["jump_server"],
        ssh_port=ssh_settings["ssh_port"],
    )

    stdin, stdout, stderr = ssh_manager.execute_command([script_cmd])
    pid = int(stdout.readline())
    return pid


def monitor_ssh(job_id: Union[list, int], ssh_settings: dict):
    """Check if PID is running on an SSH server."""
    # Create manager for file sync and subprocess exec
    ssh_manager = SSH_Manager(
        user_name=ssh_settings["user_name"],
        pkey_path=ssh_settings["pkey_path"],
        main_server=ssh_settings["main_server"],
        jump_server=ssh_settings["jump_server"],
        ssh_port=ssh_settings["ssh_port"],
    )
    # Check if PID is in running processes
    sp_script = f"ps -p {job_id}"

    while True:
        try:
            stdin, stdout, stderr = ssh_manager.execute_command([sp_script])
            break
        except Exception as e:
            print(e)
            time.sleep(0.5)

    # Check if job is listed in running ones
    counter = 0
    for line in stdout:
        counter += 1
    return counter > 1


def send_dir_ssh(
    ssh_settings: dict,
    local_dir: Union[str, None] = None,
):
    """Helper for sending code directory to SSH server."""
    if local_dir is None:
        local_dir = os.getcwd()
    # Create manager for file sync and subprocess exec
    ssh_manager = SSH_Manager(
        user_name=ssh_settings["user_name"],
        pkey_path=ssh_settings["pkey_path"],
        main_server=ssh_settings["main_server"],
        jump_server=ssh_settings["jump_server"],
        ssh_port=ssh_settings["ssh_port"],
    )
    ssh_manager.sync_dir(local_dir, ssh_settings["remote_dir"])
    return


def copy_dir_ssh(
    ssh_settings: dict,
    remote_dir: str,
    local_dir: Union[str, None] = None,
):
    """Helper for copying results directory back to local machine."""
    # Create manager for file sync and subprocess exec
    ssh_manager = SSH_Manager(
        user_name=ssh_settings["user_name"],
        pkey_path=ssh_settings["pkey_path"],
        main_server=ssh_settings["main_server"],
        jump_server=ssh_settings["jump_server"],
        ssh_port=ssh_settings["ssh_port"],
    )
    if local_dir is None:
        ssh_manager.get_file(remote_dir, os.getcwd())
    else:
        ssh_manager.get_file(remote_dir, local_dir)
    return


def delete_dir_ssh(ssh_settings: dict):
    """Delete log.txt file."""
    # Create manager for file sync and subprocess exec
    ssh_manager = SSH_Manager(
        user_name=ssh_settings["user_name"],
        pkey_path=ssh_settings["pkey_path"],
        main_server=ssh_settings["main_server"],
        jump_server=ssh_settings["jump_server"],
        ssh_port=ssh_settings["ssh_port"],
    )
    ssh_manager.delete_dir(ssh_settings["remote_dir"])



import os
import logging
import time
from typing import Union, List
import numpy as np
from rich.logging import RichHandler
from rich.progress import (
    BarColumn,
    Progress,
    TextColumn,
    TimeElapsedColumn,
    SpinnerColumn,
)
from mle_scheduler.job import MLEJob
from mle_scheduler.ssh import send_dir_ssh, copy_dir_ssh, delete_dir_ssh
from mle_scheduler.cloud.gcp import send_dir_gcp, copy_dir_gcp, delete_dir_gcp


class MLEQueue(object):
    """
    Multi-Seed Job Class: This is essentially a for-loop handler for
    the `Job` class when running multiple random seeds.
    """

    def __init__(
        self,
        resource_to_run: str,
        job_filename: str,
        job_arguments: dict = {},
        config_filenames: Union[str, List[str], None] = None,
        experiment_dir: Union[str, None] = None,
        num_seeds: int = 1,
        default_seed: int = 0,
        random_seeds: Union[None, List[int]] = None,
        max_running_jobs: int = 10,
        automerge_seeds: bool = False,
        automerge_configs: bool = False,
        delete_config: bool = False,
        debug_mode: bool = False,
        cloud_settings: Union[dict, None] = None,
        ssh_settings: Union[dict, None] = None,
        use_slack_bot: bool = False,
        slack_message_id: Union[str, None] = None,
        slack_user_name: Union[str, None] = None,
        slack_auth_token: Union[str, None] = None,
        protocol_db=None,
        logger_level: int = logging.WARNING,
    ):
        # Init experiment class with relevant info
        self.resource_to_run = resource_to_run  # compute resource for job
        self.job_filename = job_filename  # path to train script
        self.config_filenames = config_filenames  # path to config json
        self.experiment_dir = experiment_dir  # main results dir (create)
        self.job_arguments = job_arguments.copy()  # job-specific args
        self.num_seeds = num_seeds  # number seeds to run
        self.max_running_jobs = max_running_jobs  # number of sim running jobs
        self.delete_config = (
            delete_config  # Option to delete config file after run
        )
        self.debug_mode = debug_mode  # Pipe stdout and stderr to files

        # Slack Clusterbot Configuration & Protocol DB
        self.use_slack_bot = use_slack_bot  # Boolean whether to use slack bot
        self.slack_message_id = slack_message_id  # Message ts id for slack bot
        self.slack_user_name = (
            slack_user_name  # Slack user name to send message to
        )
        self.slack_auth_token = slack_auth_token  # Slack Authentication Token
        self.protocol_db = protocol_db

        # Virtual environment usage & GCS code directory/SSH settings
        self.cloud_settings = cloud_settings
        self.ssh_settings = ssh_settings

        # Time between launching jobs
        if "time_between_launch" in self.job_arguments.keys():
            self.time_between_launches = self.job_arguments["time_between_launch"]
            del self.job_arguments["time_between_launch"]
        else:
            self.time_between_launches = 0.1

        # Instantiate/connect a logger
        FORMAT = "%(message)s"
        logging.basicConfig(
            level=logger_level,
            format=FORMAT,
            datefmt="[%X]",
            handlers=[RichHandler()],
        )

        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logger_level)

        if resource_to_run == "ssh-node":
            if self.ssh_settings["start_up_copy_dir"]:
                send_dir_ssh(self.ssh_settings)
                self.logger.info("Copied code directory to SSH server")

        if resource_to_run == "gcp-cloud":
            if self.cloud_settings["start_up_copy_dir"]:
                send_dir_gcp(self.cloud_settings)
                self.logger.info("Copied code directory to GCS bucket")

        # Check whether enough seeds explicitly supplied
        if random_seeds is not None:
            self.random_seeds = random_seeds
            self.num_seeds = len(self.random_seeds)

        # Generate a list of dictionaries with different random seed cmd input
        if self.num_seeds > 1 and random_seeds is None:
            seeds_to_sample = np.arange(100000, 999999)
            self.random_seeds = np.random.choice(
                seeds_to_sample, self.num_seeds, replace=False
            ).tolist()
        elif self.num_seeds == 1 and random_seeds is None:
            self.random_seeds = [default_seed]
        self.automerge_seeds = automerge_seeds
        self.automerge_configs = automerge_configs

        # Extract extra_cmd_line_input from job_arguments
        if self.job_arguments is not None:
            if "extra_cmd_line_input" in self.job_arguments.keys():
                self.extra_cmd_line_input = self.job_arguments[
                    "extra_cmd_line_input"
                ]
                del self.job_arguments["extra_cmd_line_input"]
            else:
                self.extra_cmd_line_input = None

        # Make sure that config_filenames is list of strings to loop over
        if type(self.config_filenames) != list:
            self.config_filenames = list(self.config_filenames)

        # Generate a list of jobs to be queued/executed on the resource
        # Recreate directory in which the results are stored - later merge
        # 3 status types: -1 - not started yet; 0 - completed, 1 - running
        self.queue, self.mle_log_dirs, self.mle_run_ids = [], [], []

        for config_fname in self.config_filenames:
            base_str = os.path.split(config_fname)[1].split(".")[0]
            # Add mle log dir to storage list
            sub_experiment_dir = os.path.join(experiment_dir, base_str)
            self.mle_log_dirs.append(sub_experiment_dir)
            self.mle_run_ids.append(base_str)
            for seed_id in self.random_seeds:
                self.queue.append(
                    {
                        "config_fname": config_fname,
                        "seed_id": seed_id,
                        "base_str": base_str,
                        "experiment_dir": sub_experiment_dir,
                        "log_dir": os.path.join(sub_experiment_dir, "/logs"),
                        "status": -1,
                        "job": None,
                        "job_id": None,
                        "merged_logs": False,
                    }
                )

        self.queue_counter = 0  # Next job to schedule
        self.num_completed_jobs = 0  # No. already completed jobs
        self.num_running_jobs = 0  # No. of currently running jobs
        self.num_total_jobs = len(self.queue)  # Queue is list of all jobs

        # If no limit of jobs is manually provided - schedule all of them
        if self.max_running_jobs is None:
            self.max_running_jobs = self.num_total_jobs

        self.logger.info(
            "Queued: {} - {} seeds x {} configs".format(
                self.resource_to_run, self.num_seeds, len(self.config_filenames)
            )
        )

    def run(self) -> None:
        """Schedule -> Monitor -> Merge individual logs."""
        # 1. Spawn 1st batch of evals until limit of allowed usage is reached
        while self.num_running_jobs < min(
            self.max_running_jobs, self.num_total_jobs
        ):
            job, job_id = self.launch(self.queue_counter)
            self.queue[self.queue_counter]["status"] = 1
            self.queue[self.queue_counter]["job"] = job
            self.queue[self.queue_counter]["job_id"] = job_id
            self.num_running_jobs += 1
            self.queue_counter += 1
            time.sleep(self.time_between_launches)

        self.logger.info(
            "Launched: {} - Set of {}/{} Jobs".format(
                self.resource_to_run, self.num_running_jobs, self.num_total_jobs
            )
        )

        # 2. Set up Progress Bar Counter of completed jobs (& slack bot)
        progress = Progress(
            SpinnerColumn(),
            TextColumn(
                f"[bold blue]MLEQueue - {self.resource_to_run} •",
                justify="left",
            ),
            TextColumn(
                "{task.completed}/{task.total} Jobs",
                justify="left",
            ),
            BarColumn(bar_width=27, style="magenta"),
            TextColumn("[progress.percentage]{task.percentage:>3.0f}% •"),
            TimeElapsedColumn(),
            TextColumn(":hourglass:", justify="right"),
        )

        if (
            self.use_slack_bot
            and self.slack_user_name is not None
            and self.slack_auth_token is not None
        ):
            try:
                from clusterbot import ClusterBot
            except ImportError:
                raise ImportError(
                    "You need to install & setup `slack-clusterbot` to "
                    "use status notifications."
                )
            logger = logging.getLogger("clusterbot")
            logger.setLevel(logging.WARNING)
            slackbot = ClusterBot(
                slack_token=self.slack_auth_token,
                user_name=self.slack_user_name,
            )
            if self.slack_message_id is None:
                self.slack_message_id = slackbot.send(
                    f":rocket: Start running {self.num_total_jobs} jobs"
                    " :rocket:\n→ Compute resource:"
                    f" `{self.resource_to_run}`\n→ Bash execution file:"
                    f" `{self.job_filename}`\n→ Config .yaml:"
                    f" `{self.config_filenames}`\n→ Seeds:"
                    f" `{self.random_seeds}`",
                    user_name=self.slack_user_name,
                )
            slackbot.init_pbar(self.num_total_jobs, ts=self.slack_message_id)

        # 3. Monitor & launch new waiting jobs when resource available
        with progress:
            task = progress.add_task("queue", total=self.num_total_jobs)
            while self.num_completed_jobs < self.num_total_jobs:
                # Once budget is fully allocated - start monitor running jobs
                # Loop over all jobs in queue - check status of prev running
                for job in self.queue:
                    if job["status"] == 1:
                        status = self.monitor(job["job"], job["job_id"], False)
                        # If status changes to completed - update counters/state
                        if status == 0:
                            self.num_completed_jobs += 1
                            self.num_running_jobs -= 1
                            job["status"] = 0
                            # Clean up after job completion (e.g VM instance)
                            if not self.debug_mode:
                                job["job"].clean_up(job["job_id"])
                            # Update the rich progress bar after job completed
                            progress.advance(task)

                            # Update the slack progress bar
                            if (
                                self.use_slack_bot
                                and self.slack_user_name is not None
                                and self.slack_auth_token is not None
                            ):
                                try:
                                    slackbot.update_pbar()
                                except Exception:
                                    pass

                            # Update the protocol db progress bar
                            if self.protocol_db is not None:
                                self.protocol_db.update_progress_bar()
                    time.sleep(0.1)

                # Once budget becomes available again - fill up with new jobs
                while self.num_running_jobs < min(
                    self.max_running_jobs,
                    self.num_total_jobs
                    - self.num_completed_jobs
                    - self.num_running_jobs,
                ):
                    job, job_id = self.launch(self.queue_counter)
                    self.queue[self.queue_counter]["status"] = 1
                    self.queue[self.queue_counter]["job"] = job
                    self.queue[self.queue_counter]["job_id"] = job_id
                    self.num_running_jobs += 1
                    self.queue_counter += 1
                    time.sleep(0.1)

        self.logger.info(
            "Completed: {} - {}/{} Jobs".format(
                self.resource_to_run, self.num_total_jobs, self.num_total_jobs
            )
        )

        if self.resource_to_run == "ssh-node":
            copy_dir_ssh(
                self.ssh_settings,
                remote_dir=os.path.join(
                    self.ssh_settings["remote_dir"], self.experiment_dir
                ),
            )
            self.logger.info(f"Pulled SSH results - {self.experiment_dir}")
            # Clean up the scp code directory
            if "clean_up_remote_dir" in self.ssh_settings.keys():
                if self.ssh_settings["clean_up_remote_dir"]:
                    delete_dir_ssh(self.ssh_settings)
                    self.logger.info(
                        "Deleted SSH directory -"
                        f" {self.ssh_settings['remote_dir']}"
                    )

        elif self.resource_to_run == "gcp-cloud":
            copy_dir_gcp(
                self.cloud_settings,
                remote_dir=os.path.join(
                    self.cloud_settings["remote_dir"], self.experiment_dir
                ),
            )
            self.logger.info(f"Pulled cloud results - {self.experiment_dir}")
            # Clean up the scp code directory
            if "clean_up_remote_dir" in self.cloud_settings.keys():
                if self.cloud_settings["clean_up_remote_dir"]:
                    delete_dir_gcp(self.cloud_settings)
                    self.logger.info(
                        "Deleted cloud directory -"
                        f" {self.cloud_settings['remote_dir']}"
                    )

        # Merge configs and/or seeds of one eval/config if all jobs done!
        if self.automerge_configs:
            self.merge_configs(merge_seeds=True)
            self.logger.info(f"Merged config logs for different directories")
        elif self.automerge_seeds:
            for log_dir in self.mle_log_dirs:
                self.merge_seeds(log_dir)
            self.logger.info(
                f"Merged seeds for log directories - {self.mle_log_dirs}"
            )

    def launch(self, queue_counter):
        """Launch a set of jobs for one configuration - one for each seed."""
        # 1. Instantiate the experiment class and start a single seed
        job = MLEJob(
            self.resource_to_run,
            self.job_filename,
            self.job_arguments,
            self.queue[queue_counter]["config_fname"],
            self.experiment_dir,
            self.queue[queue_counter]["seed_id"],
            self.extra_cmd_line_input,
            self.delete_config,
            self.debug_mode,
            self.cloud_settings,
            self.ssh_settings,
        )

        # 2. Launch a single experiment
        job_id = job.schedule()

        # 3. Return updated counter, `Job` instance & corresponding ID
        return job, job_id

    def monitor(self, job: MLEJob, job_id: str, continuous: bool = True):
        """Monitor all seed-specific jobs for one eval configuration."""
        if continuous:
            while True:
                status = job.monitor(job_id, False)
                if status == 0:
                    return 0
        else:
            status = job.monitor(job_id, False)
            return status

    def merge_seeds(self, experiment_dir: str, load: bool = True) -> None:
        """Collect all seed-specific seeds into single log.hdf5 file."""
        try:
            from mle_logging import merge_seed_logs, load_log
        except ModuleNotFoundError as err:
            raise ModuleNotFoundError(
                f"{err}. You need to install `mle_logging` "
                "to use the `merge_seeds` method."
            )

        # Only merge logs if job is based on python job!
        # Otherwise .hdf5 file system is not used and there is nothing to merge
        # filename, file_extension = os.path.splitext(self.job_filename)
        # if file_extension == ".py":
        merged_path = os.path.join(experiment_dir, "logs", "log.hdf5")
        merge_seed_logs(merged_path, experiment_dir, self.num_seeds)
        if load:
            self.log = load_log(experiment_dir)

    def merge_configs(self, merge_seeds: bool = False, load: bool = True):
        """Collect all config-specific logs into single meta_log.hdf5 file."""
        try:
            from mle_logging import merge_config_logs, load_log

        except ModuleNotFoundError as err:
            raise ModuleNotFoundError(
                f"{err}. You need to install `mle_logging` "
                "to use the `merge_configs` method."
            )
        if merge_seeds:
            for log_dir in self.mle_log_dirs:
                self.merge_seeds(log_dir)
        merge_config_logs(
            experiment_dir=self.experiment_dir, all_run_ids=self.mle_run_ids
        )

        if load:
            self.log = load_log(
                os.path.join(self.experiment_dir, "meta_log.hdf5"),
                aggregate_seeds=True,
            )



try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup, find_packages

import re
import os
from typing import List

CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(CURRENT_DIR, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


def parse_requirements(path: str) -> List[str]:
    with open(os.path.join(CURRENT_DIR, path)) as f:
        return [
            line.rstrip()
            for line in f
            if not (line.isspace() or line.startswith("#"))
        ]


VERSIONFILE = "mle_scheduler/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))
git_tar = f"https://github.com/mle-infrastructure/mle-scheduler/archive/v{verstr}.tar.gz"


setup(
    name="mle_scheduler",
    version=verstr,
    author="Robert Tjarko Lange",
    author_email="robertlange0@gmail.com",
    description="Machine Learning Experiment Job Scheduler",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mle-infrastructure/mle-scheduler",
    download_url=git_tar,
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    platforms="any",
    python_requires=">=3.6",
    install_requires=parse_requirements(
        os.path.join(CURRENT_DIR, "requirements", "requirements.txt")
    ),
    tests_require=parse_requirements(
        os.path.join(CURRENT_DIR, "requirements", "requirements-test.txt")
    ),
    extras_require={
        "examples": parse_requirements(
            os.path.join(
                CURRENT_DIR, "requirements", "requirements-examples.txt"
            )
        ),
        "full": ["google-cloud-storage", "paramiko", "sshtunnel", "scp"],
    },
)
